package onebit.chrdraw.chr;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

public class FontFactory {

  static byte[] SIGNATURE = new byte[] { 
          'P', 'K', 8, 8 
      };

  public static Font load(InputStream is) throws IOException {
    DataInputStream dis = new DataInputStream(is);
    // header
    readSignature(dis);
    String fontType = readString(dis, 4);
    String description = readDescription(dis);
    int headerSize = readUnsignedShort(dis);
    String internalFontname = readString(dis, 4);
    int fileSize = readUnsignedShort(dis);
    int driverMajorVersion = dis.readUnsignedByte();
    int driverMinorVersion = dis.readUnsignedByte();
    readMarker(dis);
    readPad(dis, headerSize 
        - SIGNATURE.length - fontType.length() - description.length() 
        - 1 - 2
        - internalFontname.length() 
        - 2 - 1 - 1 - 2);

    char signature = (char) dis.readUnsignedByte();
    int numberOfChars = readUnsignedShort(dis);
    dis.readUnsignedByte();
    char firstChar = (char) dis.readUnsignedByte();
    int offsetToStrokeDefinitions = readUnsignedShort(dis);
    int scanFlag = dis.readUnsignedByte();
    int originToTopOfCapital = dis.readUnsignedByte();
    int originToBaseLine = dis.readUnsignedByte();
    int originToBottomDescender = dis.readUnsignedByte();
    String fontname = readString(dis, 4);
    dis.readUnsignedByte();

    int[] offsetsToCharacterDefinitions = readOffsetsToCharacterDefinitions(
        dis, numberOfChars);
    int[] widthsOfCharacters = readWidthsOfCharacters(dis, numberOfChars);
    Map<Character, Glyph> characterDefinitions = readGlyphs(dis, 
        firstChar, widthsOfCharacters, offsetsToCharacterDefinitions, 
        fileSize - offsetToStrokeDefinitions);

    return new Font(fontType, description, internalFontname, 
        driverMajorVersion, driverMinorVersion, signature, scanFlag, fontname,
        originToTopOfCapital, originToBaseLine, originToBottomDescender,
        characterDefinitions);
  }

  static void readSignature(DataInputStream dis) throws IOException {
    byte[] signature = new byte[SIGNATURE.length];
    dis.read(signature);
    if (!Arrays.equals(SIGNATURE, signature)) {
      throw new IOException("Bad CHR signature");
    }
  }

  static String readDescription(DataInputStream dis) throws IOException {
    StringBuffer buf = new StringBuffer();
    int ch;
    while ((ch = dis.readUnsignedByte()) != 26) {
      buf.append((char) ch);
    }
    return buf.toString();
  }

  static String readString(DataInputStream dis, int length) throws IOException {
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < length; i++) {
      int ch = dis.readUnsignedByte();
      buf.append((char) ch);
    }
    return buf.toString();
  }

  static void readMarker(DataInputStream dis) throws IOException {
    int marker = readUnsignedShort(dis);
    if (marker != 1) {  //odwrócona kolejność bajtów 
      throw new IOException("Incorrect marker 0x100");
    }
  }

  static void readPad(DataInputStream dis, int length) throws IOException {
    for (int i = 0; i < length; i++) {
      dis.readUnsignedByte();
    }
  }

  static int[] readOffsetsToCharacterDefinitions(DataInputStream dis, int length)
      throws IOException {
    int[] offsets = new int[length];
    for (int i = 0; i < length; i++) {
      int offset = readUnsignedShort(dis);
      offsets[i] = offset;
    }
    return offsets;
  }

  static int[] readWidthsOfCharacters(DataInputStream dis, int length)
      throws IOException {
    int[] widths = new int[length];
    for (int i = 0; i < length; i++) {
      int width = dis.readUnsignedByte();
      widths[i] = width;
    }
    return widths;
  }

  static Map<Character, Glyph> readGlyphs(DataInputStream dis, char firstChar,
      int[] widthsOfCharacters, int[] offsetsToCharacterDefinitions,
      int offsetToEndOfDefinitions) throws IOException {
    Map<Character, Glyph> characterDefinitions = new HashMap<Character, Glyph>();
    for (int i = 0; i < offsetsToCharacterDefinitions.length; i++) {
      char ch = (char)(firstChar + i);
      int width = widthsOfCharacters[i];
      int length = ( ( i < offsetsToCharacterDefinitions.length - 1? 
            offsetsToCharacterDefinitions[i + 1]: 
            offsetToEndOfDefinitions ) 
            - offsetsToCharacterDefinitions[i] ) / 2;
      Collection<Command> commands = readCommands(dis, length);
      Glyph glyph = new Glyph(ch, width, commands);
      characterDefinitions.put(ch, glyph);
    }
    return characterDefinitions;
  }

  static Collection<Command> readCommands(DataInputStream dis, int length) throws IOException {
    Collection<Command> commands = new ArrayList<Command>();
    for (int i = 0; i < length; i++) {
      int byte1 = dis.readUnsignedByte();
      int byte2 = dis.readUnsignedByte();
      int x = shrinkCoordinate(byte1);
      int y = shrinkCoordinate(byte2);
      Opcode opcode = makeOpcode(byte1, byte2);
      Command command = new Command(opcode, x, y);
      commands.add(command);
    }
    return commands;
  }
  
  static int shrinkCoordinate(int data) {
    return ((byte)(data << 1)) >> 1;  //rozszerzenie znaku zapisanego na bicie 6
  }
  
  static Opcode makeOpcode(int byte1, int byte2) {
    return Opcode.values()[((byte1 & 0x80) | (byte2 >>> 1)) >>> 6];
  }
  
  static int readUnsignedShort(DataInputStream dis) throws IOException {
    int byte1 = dis.readUnsignedByte();
    int byte2 = dis.readUnsignedByte();
    return byte2 << 8 | byte1;
  }
  
}
